#include <boost/program_options.hpp>

#include <iostream>
#include <thread>
#include <mutex>

#include <boost/beast/http.hpp>
#include <boost/beast/core/handler_ptr.hpp>
#include <boost/beast/core/multi_buffer.hpp>
#include <boost/beast/core/flat_buffer.hpp>
#include <boost/asio.hpp>

#include <app_args.hpp>

using namespace boost::beast;
using namespace boost::beast::http;

template < typename Strand >
class http_async_server
{
    using endpoint_type = boost::asio::ip::tcp::endpoint;
    using address_type = boost::asio::ip::address;
    using socket_type = boost::asio::ip::tcp::socket;

    using req_type = request<string_body>;
    using resp_type = response<string_body>;

    std::mutex m_;
    bool log_ = true;
    boost::asio::io_service ios_;
    boost::asio::ip::tcp::acceptor acceptor_;
    socket_type sock_;
    std::string root_;
    std::vector<std::thread> thread_;

public:
    http_async_server(endpoint_type const& ep,
            std::size_t threads)
        : acceptor_(ios_)
        , sock_(ios_)
    {
        acceptor_.open(ep.protocol());
        acceptor_.bind(ep);
        acceptor_.listen(
            boost::asio::socket_base::max_connections);
        acceptor_.async_accept(sock_,
            std::bind(&http_async_server::on_accept, this,
                std::placeholders::_1));
        thread_.reserve(threads);
        for(std::size_t i = 0; i < threads; ++i)
            thread_.emplace_back(
                [&] { ios_.run(); });
    }

    ~http_async_server()
    {
        error_code ec;
        ios_.dispatch(
            [&]{ acceptor_.close(ec); });
        for(auto& t : thread_)
            t.join();
    }

private:
    template<class Stream, class Handler,
        bool isRequest, class Body, class Fields>
    class write_op
    {
        struct data
        {
            bool cont;
            Stream& s;
            message<isRequest, Body, Fields> m;

            data(Handler& handler, Stream& s_,
                    message<isRequest, Body, Fields>&& m_)
                : s(s_)
                , m(std::move(m_))
            {
                using boost::asio::asio_handler_is_continuation;
                cont = asio_handler_is_continuation(std::addressof(handler));
            }
        };

        handler_ptr<data, Handler> d_;

    public:
        write_op(write_op&&) = default;
        write_op(write_op const&) = default;

        template<class DeducedHandler, class... Args>
        write_op(DeducedHandler&& h, Stream& s, Args&&... args)
            : d_(std::forward<DeducedHandler>(h),
                s, std::forward<Args>(args)...)
        {
            (*this)(error_code{}, false);
        }

        void
        operator()(error_code ec, bool again = true)
        {
            auto& d = *d_;
            d.cont = d.cont || again;
            if(! again)
            {
                boost::beast::http::async_write(d.s, d.m, std::move(*this));
                return;
            }
            d_.invoke(ec);
        }

        friend
        void* asio_handler_allocate(
            std::size_t size, write_op* op)
        {
            using boost::asio::asio_handler_allocate;
            return asio_handler_allocate(
                size, std::addressof(op->d_.handler()));
        }

        friend
        void asio_handler_deallocate(
            void* p, std::size_t size, write_op* op)
        {
            using boost::asio::asio_handler_deallocate;
            asio_handler_deallocate(
                p, size, std::addressof(op->d_.handler()));
        }

        friend
        bool asio_handler_is_continuation(write_op* op)
        {
            return op->d_->cont;
        }

        template<class Function>
        friend
        void asio_handler_invoke(Function&& f, write_op* op)
        {
            using boost::asio::asio_handler_invoke;
            asio_handler_invoke(
                f, std::addressof(op->d_.handler()));
        }
    };

    template<class Stream,
        bool isRequest, class Body, class Fields,
            class DeducedHandler>
    static
    void
    async_write(Stream& stream, message<
        isRequest, Body, Fields>&& msg,
            DeducedHandler&& handler)
    {
        write_op<Stream, typename std::decay<DeducedHandler>::type,
            isRequest, Body, Fields>{std::forward<DeducedHandler>(
                handler), stream, std::move(msg)};
    }

    class peer : public std::enable_shared_from_this<peer>
    {
        int id_;
        flat_buffer sb_{ 1024 };
        socket_type sock_;
        http_async_server& server_;
        Strand strand_;
        req_type req_;

    public:
        peer(peer&&) = default;
        peer(peer const&) = default;
        peer& operator=(peer&&) = delete;
        peer& operator=(peer const&) = delete;

        peer(socket_type&& sock, http_async_server& server)
            : sock_(std::move(sock))
            , server_(server)
            , strand_(sock_.get_io_service())
        {
            static int n = 0;
            id_ = ++n;
        }

        void
        fail(error_code ec, std::string what)
        {
        }

        void run()
        {
            do_read();
        }

        void do_read()
        {
            async_read(sock_, sb_, req_,
                strand_.wrap(
                std::bind(&peer::on_read, this->shared_from_this(),
                    std::placeholders::_1)
                ) );
        }

        void on_read(error_code const& ec)
        {
            if(ec)
                return fail(ec, "read");

            resp_type res;
            res.result(status::ok);
            res.version( req_.version() );
            res.insert(field::server, "Beast benchmark");
            res.insert(field::content_type, "text/plain; charset=utf-8");
            res.keep_alive( true );
            res.body() = "Hello world!";
            res.prepare_payload();
            async_write(sock_, std::move(res),
                strand_.wrap(
                    std::bind(&peer::on_write, this->shared_from_this(),
                        std::placeholders::_1)));
        }

        void on_write(error_code ec)
        {
            if(!ec)
                do_read();
        }
    };

    void
    on_accept(error_code ec)
    {
        if(! acceptor_.is_open())
            return;

        if(!ec)
        {
            socket_type sock(std::move(sock_));
            std::make_shared<peer>(std::move(sock), *this)->run();
        }

        acceptor_.async_accept(sock_,
            std::bind(&http_async_server::on_accept, this,
                std::placeholders::_1));
    }
};

using http_async_server_st = http_async_server< boost::asio::io_service & >;
using http_async_server_mt = http_async_server< boost::asio::io_service::strand >;

using endpoint_type = boost::asio::ip::tcp::endpoint;
using address_type = boost::asio::ip::address;

template < typename Server >
void run_server( endpoint_type ep, std::size_t pool_size )
{
    Server server{ ep, pool_size };

    std::cout << "Press ENTER to exit." << std::endl;
    std::string line;
    std::getline( std::cin, line );
}

int main(int ac, char const* av[])
{
    try
    {
        const app_args_t args{ ac, av };

        if( !args.m_help )
        {
            std::cout << "pool size: " << args.m_pool_size << std::endl;

            endpoint_type ep{address_type::from_string( "127.0.0.1" ), args.m_port };

            if( 1 == args.m_pool_size )
            {
                run_server< http_async_server_st >( ep, args.m_pool_size );
            }
            else if ( 1 < args.m_pool_size )
            {
                run_server< http_async_server_mt >( ep, args.m_pool_size );
            }
            else
            {
                throw std::runtime_error{ "invalid asio pool size" };
            }
        }
    }
    catch( const std::exception & ex )
    {
        std::cerr << "Error: " << ex.what() << std::endl;
        return 1;
    }

    return 0;
}
